from splatnet3_scraper.query.config import Config
from splatnet3_scraper.query.handler import QueryHandler
from splatnet3_scraper.query.json_parser import JSONParser
from splatnet3_scraper.query.responses import QueryResponse



from typing import Literal, cast, overload

from splatnet3_scraper.query import QueryHandler, QueryResponse
from splatnet3_scraper.scraper.query_map import QueryMap


class SplatNet_Scraper:
    """This class offers a user-level interface for pulling data from SplatNet
    3. It is built upon the QueryHandler class and provides a top-level
    API that orchestrates multiple queries together to reduce the amount of work
    needed to pull data that users are likely to want.
    """

    def __init__(self, query_handler: QueryHandler) -> None:
        """Initializes a SplatNet_Scraper.

        Args:
            query_handler (QueryHandler): The query handler to use.
        """
        self._query_handler = query_handler

    @staticmethod
    def from_session_token(session_token: str) -> "SplatNet_Scraper":
        """Creates a SplatNet_Scraper instance using the given session token.

        Args:
            session_token (str): The session token to use.

        Returns:
            SplatNet_Scraper: The SplatNet_Scraper instance.
        """
        query_handler = QueryHandler.from_session_token(session_token)
        return SplatNet_Scraper(query_handler)

    @staticmethod
    def from_config_file(config_path: str | None = None) -> "SplatNet_Scraper":
        """Creates a SplatNet_Scraper instance using the given config file.

        Args:
            config_path (str | None): The path to the config file. If None, it
                will look for ".splatnet3_scraper" in the current working
                directory.

        Returns:
            SplatNet_Scraper: The SplatNet_Scraper instance.
        """
        query_handler = QueryHandler.from_config_file(config_path)
        return SplatNet_Scraper(query_handler)

    @staticmethod
    def from_env() -> "SplatNet_Scraper":
        """Creates a SplatNet_Scraper instance using the environment variables.

        Environment variables:
            SN3S_SESSION_TOKEN: The session token to use.
            SN3S_GTOKEN: The gtoken to use.
            SN3S_BULLET_TOKEN: The bullet token to use.

        Returns:
            SplatNet_Scraper: The SplatNet_Scraper instance.
        """
        query_handler = QueryHandler.from_env()
        return SplatNet_Scraper(query_handler)

    @staticmethod
    def from_s3s_config(config_path: str) -> "SplatNet_Scraper":
        """Creates a SplatNet_Scraper instance using the config file from s3s.

        Args:
            config_path (str): The path to the config file.

        Returns:
            SplatNet_Scraper: The SplatNet_Scraper instance.
        """
        query_handler = QueryHandler.from_s3s_config(config_path)
        return SplatNet_Scraper(query_handler)

    def __query(self, query: str, variables: dict = {}) -> QueryResponse:
        """Convenience function for querying.

        Args:
            query (str): The query to run.
            variables (dict): The variables to pass to the query. Defaults to
                an empty dict.

        Returns:
            QueryResponse: The QueryResponse.
        """
        return self._query_handler.query(query, variables)

    def __detailed_vs_or_coop(
        self,
        query: str,
        limit: int | None = None,
        existing_ids: list[str] | str | None = None,
    ) -> tuple[QueryResponse, list[QueryResponse]]:
        """Gets the detailed results for a vs battle or coop battle.

        Args:
            query (str): The query to run.
            limit (int | None): The maximum number of battles to get. If None,
                it will get all battles. Defaults to None.
            existing_ids (list[str] | str | None): The existing IDs to check
                against. If a string is passed, it will return the results
                upon finding the first match. If a list is passed, it will
                return the results of all matches not in the list. If None,
                it will return all results. Defaults to None.

        Raises:
            ValueError: If the query is not a vs battle or coop battle.

        Returns:
            tuple:
                QueryResponse: The summary query response.
                list[QueryResponse]: The list of detailed query responses
                    associated with each battle until the limit is reached.
        """
        if query not in (
            QueryMap.SALMON,
            QueryMap.TURF,
            QueryMap.ANARCHY,
            QueryMap.XBATTLE,
            QueryMap.PRIVATE,
        ):
            raise ValueError(f"Invalid query: {query}")

        if query in (QueryMap.SALMON, QueryMap.SALMON_DETAIL):
            detail_query = QueryMap.SALMON_DETAIL
            variable_name = "coopHistoryDetailId"
        else:
            detail_query = QueryMap.VS_DETAIL
            variable_name = "vsResultId"

        _limit = -1 if limit is None else limit

        # Get the list of battles
        summary_query = self.__query(query)

        # Top level key depends on the game mode, but there is only one.
        top_level_key = summary_query.keys()[0]
        history_groups = summary_query[top_level_key, "historyGroups", "nodes"]
        out: list[QueryResponse] = []
        idx = 0

        for group in history_groups:
            group = cast(QueryResponse, group)
            for game in group["historyDetails", "nodes"]:
                game = cast(QueryResponse, game)
                if idx == _limit:
                    return summary_query, out
                idx += 1
                game_id = game["id"]

                if isinstance(existing_ids, str):
                    if game_id == existing_ids:
                        return summary_query, out
                elif isinstance(existing_ids, list):
                    if game_id in existing_ids:
                        idx -= 1
                        continue

                variables = {variable_name: game_id}
                detailed_game = self.__query(detail_query, variables)
                out.append(detailed_game)

        return summary_query, out

    @overload
    def get_vs_battles(
        self,
        mode: str,
        detail: Literal[False],
        limit: int | None = None,
        existing_ids: list[str] | str | None = None,
    ) -> QueryResponse:
        ...

    @overload
    def get_vs_battles(
        self,
        mode: str,
        detail: Literal[True],
        limit: int | None = None,
        existing_ids: list[str] | str | None = None,
    ) -> tuple[QueryResponse, list[QueryResponse]]:
        ...

    @overload
    def get_vs_battles(
        self,
        mode: str,
        detail: bool = False,
        limit: int | None = None,
        existing_ids: list[str] | str | None = None,
    ) -> QueryResponse | tuple[QueryResponse, list[QueryResponse]]:
        ...

    def get_vs_battles(
        self,
        mode: str,
        detail: bool = False,
        limit: int | None = None,
        existing_ids: list[str] | str | None = None,
    ) -> QueryResponse | tuple[QueryResponse, list[QueryResponse]]:
        """Gets the vs battles.

        Args:
            mode (str): The mode to get the battles for. Some values are:
                "turf", "anarchy", "xbattle", "private",
            detail (bool): Whether to get the detailed results or not.
                Defaults to False.
            limit (int | None): The maximum number of battles to get. If None,
                it will get all battles. Defaults to None.
            existing_ids (list[str] | str | None): The existing IDs to check
                against. If a string is passed, it will return the results
                upon finding the first match. If a list is passed, it will
                return the results of all matches not in the list. If None,
                it will return all results. Defaults to None.

        Raises:
            ValueError: If the mode is not valid.

        Returns:
            QueryResponse : The summary query response, returned regardless of
                the value of detail.
            (list[QueryResponse]): The list of detailed query responses
                associated with each battle until the limit is reached. Only
                returned if detail is True, along with the summary query
                response.
        """
        mapped_query = QueryMap.get(mode)
        if mapped_query == QueryMap.SALMON:
            raise ValueError("Use get_coop_battles for salmon run battles.")

        if mapped_query in (
            QueryMap.TURF_DETAIL,
            QueryMap.ANARCHY_DETAIL,
            QueryMap.XBATTLE_DETAIL,
            QueryMap.PRIVATE_DETAIL,
        ):
            non_detail_map = {
                QueryMap.TURF_DETAIL: QueryMap.TURF,
                QueryMap.ANARCHY_DETAIL: QueryMap.ANARCHY,
                QueryMap.XBATTLE_DETAIL: QueryMap.XBATTLE,
                QueryMap.PRIVATE_DETAIL: QueryMap.PRIVATE,
            }
            mapped_query = non_detail_map[mapped_query]
            detail = True

        if mapped_query not in (
            QueryMap.TURF,
            QueryMap.ANARCHY,
            QueryMap.XBATTLE,
            QueryMap.PRIVATE,
        ):
            raise ValueError(f"Invalid mode: {mode}")

        if detail:
            return self.__detailed_vs_or_coop(mapped_query, limit, existing_ids)
        else:
            return self.__query(mapped_query)



import configparser
import json
import os
import re
import time
from typing import Any, Literal, cast, overload

import requests

from splatnet3_scraper import __version__
from splatnet3_scraper.auth.environment_manager import (
    EnvironmentVariablesManager,
)
from splatnet3_scraper.auth.exceptions import (
    NintendoException,
    SplatNetException,
)
from splatnet3_scraper.auth.graph_ql_queries import queries
from splatnet3_scraper.auth.nso import NSO
from splatnet3_scraper.constants import (
    ENV_VAR_NAMES,
    GRAPH_QL_REFERENCE_URL,
    TOKEN_EXPIRATIONS,
    TOKENS,
)
from splatnet3_scraper.utils import retry

text_config_re = re.compile(r"\s*=*\s*")


class Token:
    """Class that represents a token. This class is meant to store the token
    itself, the type of token it is, and the time it was created. It can be used
    to check if the token is expired or display the time left before it expires.
    It also provides convenience methods for getting all sorts of metadata about
    the token.
    """

    def __init__(self, token: str, token_type: str, timestamp: float) -> None:
        """Initializes a ``Token`` object. The expiration time is calculated
        based on the token type, with a default of ``1e10`` seconds (about 316
        days, this should be basically forever for all intents and purposes; if
        you have a python session running for that long, you have bigger
        problems than a token expiring).

        Args:
            token (str): The value of the token, this is the actual token.
            token_type (str): The type of token, this is used to identify which
                type of token it represents, making it easier for the manager
                to handle the tokens when searching for a specific one. It also
                determines the expiration time of the token.
            timestamp (float): The time the token was created, in seconds since
                the epoch. This is used to determine if the token is expired.
        """
        self.token = token
        self.token_type = token_type
        self.timestamp = timestamp
        self.expiration = TOKEN_EXPIRATIONS.get(token_type, 1e10) + timestamp

    @property
    def is_valid(self) -> bool:
        """A very rudimentary check to see if the token is valid. This is not
        a guarantee that the token is valid, but it is a good indicator that
        it is for most cases. It checks if the token is not None and if it is
        not an empty string. This usually means that the token is valid, but
        it is not a guarantee. This is also here in case a future version of
        the API requires a different check to determine if a token is valid.

        Returns:
            bool: True if the token is valid (not None and not an empty string)
                False otherwise.
        """
        return (self.token is not None) and (self.token != "")

    @property
    def is_expired(self) -> bool:
        """Checks if the token is expired. This is done by comparing the
        current time to the expiration time of the token. If the current time
        is greater than the expiration time, the token is expired. This is not
        a guarantee that the token is expired, but it is a good indicator that
        it is for most cases.

        Returns:
            bool: True if the token is expired, False otherwise.
        """
        return self.time_left <= 0

    @property
    def time_left(self) -> float:
        """Returns the time left before the token expires. If the token is
        expired, a negative number will be returned. This is not a guarantee
        that the token is expired, but it is a good indicator that it is for
        most cases.

        Returns:
            float: The time left before the token expires.
        """
        return self.expiration - time.time()

    @property
    def time_left_str(self) -> str:
        """A string representation of the time left before the token expires.
        If the token is expired, "Expired" will be returned. This is not a
        guarantee that the token is expired, but it is a good indicator that
        it is for most cases. If the time left is greater than 100,000 hours,
        "basically forever" will be returned. If you have a python session
        running for that long, you have bigger problems than a token expiring.

        Returns:
            str: A string representation of the time left before the token
                expires.
        """
        time_left = self.time_left
        if time_left <= 0:
            return "Expired"
        mins, secs = divmod(time_left, 60)
        hours, mins = divmod(mins, 60)

        out = ""
        if hours > 1e5:
            return "basically forever"
        if hours > 0:
            out += f"{hours:.0f}h "
        if mins > 0:
            out += f"{mins:.0f}m "
        if secs > 0:
            out += f"{secs:.1f}s"
        return out.strip()

    def __repr__(self) -> str:
        out = "Token("
        spaces = " " * len(out)
        out += (
            f"token={self.token[:5]}...,\n"
            + spaces
            + f"type={self.token_type},\n"
            + spaces
            + "expires in "
            + self.time_left_str
            + "\n)"
        )
        return out


class TokenManager:
    """Class that manages tokens. This class can be used to add tokens, generate
    tokens from the ``NSO`` class, check if tokens are expired, load tokens from
    a configuration file or environment variables, save tokens to disk, and
    display the time left before tokens expire among other things.
    """

    def __init__(
        self,
        nso: NSO | None = None,
        *args,
        env_manager: EnvironmentVariablesManager | None = None,
    ) -> None:
        """Initializes a TokenManager object. This will create a new instance
        of the NSO class if one is not provided.

        Args:
            nso (NSO | None, optional): An instance of the NSO class. If one is
                not provided, a new instance will be created. Defaults to None.
            *args: Additional arguments to pass to the NSO class, currently not
                used.
            env_manager (EnvironmentVariablesManager): An instance of the
                EnvironmentVariablesManager class or a subclass of it. If not
                provided, a new instance will be created.
        """
        nso = nso if nso is not None else NSO.new_instance()
        self.nso = nso
        self._tokens: dict[str, Token] = {}
        self._data: dict[str, str] = {}
        self.env_manager = (
            env_manager
            if env_manager is not None
            else EnvironmentVariablesManager()
        )

    def flag_origin(self, origin: str, data: str | None = None) -> None:
        """Flags the origin of the token manager. This is used to identify where
        the token manager was loaded from, if anywhere. This is used to help
        keep track of whether the user wants to save the tokens to disk or not,
        but can potentially be used for other things in the future. This is
        called automatically when the token manager is loaded from a config
        file or environment variables. Subsequent calls to this method will
        overwrite the previous origin.

        Args:
            origin (str): The origin of the token manager.
            data (str | None): Additional data about the origin. For example,
                if the token manager was loaded from a config file, this would
                be the path to the config file. On the other hand, if the token
                manager was loaded from environment variables, this would be
                None.
        """
        self._origin = {"origin": origin, "data": data}

    @property
    def origin(self) -> dict[str, str | None]:
        """Gets the origin of the token manager.

        This method is a thin wrapper around the ``_origin`` attribute. If the
        attribute does not exist, this method will return an empty dictionary
        but will not create the attribute. The attribute will only be created
        when the ``flag_origin`` method is called.

        Returns:
            dict[str, str | None]: The origin of the token manager. This will
                be a dictionary with two keys: ``origin`` and ``data``. The
                ``origin`` key will contain a string that describes where the
                token manager was loaded from. The ``data`` key will contain
                additional data about the origin. For example, if the token
                manager was loaded from a config file, this would be the path
                to the config file. On the other hand, if the token manager was
                loaded from environment variables, this would be None.
        """
        return getattr(self, "_origin", {})

    def add_token(
        self,
        token: str | Token,
        token_type: str | None = None,
        timestamp: float | None = None,
    ) -> None:
        """Adds a token to the manager. If the token provided is a ``string``,
        the token type must be provided as well. The manager will then create a
        new Token object and add it to its internal dictionary. If the token
        provided is a ``Token`` object, the object will be added to the manager
        as is.

        If the token type is ``session_token``, the token will

        Args:
            token (str | Token): The token to add to the manager. If a string,
                the token type must be provided as well. If a ``Token`` object,
                the token type will be ignored. This field might also be known
                as the "access token" or "auth token" among other names.
            token_type (str | None): The type of token. If token is an instance
                of ``Token``, this will be ignored. If token is a string, this
                must be provided. Currently the only supported token types that
                are recognized by the manager are "session_token", "gtoken", and
                "bullet_token". Defaults to None, which will cause an error to
                be raised if ``token`` is a string.
            timestamp (float | None): The time the token was created. If not
                provided, the current time will be used.

        Raises:
            ValueError: If the ``token`` provided is a string and ``token_type``
                is not provided.
        """
        if timestamp is None:
            timestamp = time.time()
        if isinstance(token, str) and token_type is None:
            raise ValueError(
                "arg token_type must be provided if token is a str."
            )
        elif isinstance(token, str) and token_type is not None:
            token_object = Token(token, token_type, timestamp)
        elif isinstance(token, Token):
            token_object = token

        self._tokens[token_object.token_type] = token_object
        if token_object.token_type == TOKENS.GTOKEN:
            self.nso._gtoken = token_object.token
        elif token_object.token_type == TOKENS.SESSION_TOKEN:
            self.nso._session_token = token_object.token

        return

    @overload
    def get(self, token_type: str, full_token: Literal[False] = ...) -> str:
        ...

    @overload
    def get(self, token_type: str, full_token: Literal[True]) -> Token:
        ...

    @overload
    def get(self, token_type: str, full_token: bool) -> str | Token:
        ...

    def get(self, token_type: str, full_token: bool = False) -> str | Token:
        """Gets a token from the manager given a token type.

        If ``full_token`` is True, the full ``Token`` object will be returned.
        Otherwise, just the value of the token will be returned. If the token
        type is not found, a ValueError will be raised.

        Args:
            token_type (str): The type of the token to get, as defined in the
                ``Token.token_type`` field.
            full_token (bool): Whether to return the full ``Token`` object or
                just the value of the token. Defaults to False.

        Raises:
            ValueError: If the given token type is not found in the manager.

        Returns:
            str | Token: The token, either as a string or a ``Token`` object as
                defined by the ``full_token`` argument.
        """
        token_obj = self._tokens.get(token_type, None)
        if token_obj is None:
            raise ValueError(f"Token of type {token_type} not found.")
        if full_token:
            return token_obj
        return token_obj.token

    @property
    def data(self) -> dict[str, Any]:
        """Returns the full data stored in the manager as a dictionary. This
        data is not the same as the tokens stored in the manager, but rather
        additional data that can be used in other places. Examples of keys that
        might be stored in this dictionary are the user's ``country``, and
        ``language`` as defined by the Nintendo Account.

        Returns:
            dict[str, Any]: The data stored in the manager as a dictionary.
        """
        return self._data

    def add_session_token(self, token: str) -> None:
        """Adds a session token to the manager. This is a convenience method
        that will also set the session token in the NSO class. This
        functionality is already present in the ``add_token`` method, so this
        method might be removed in the future.

        Args:
            token (str): The session token to add to the manager. This field
                might also be known as the "access token" or "auth token" among
                other names. This token is used to authenticate with Nintendo
                services, and is required to query the SplatNet 3 API.
        """
        self.add_token(token, TOKENS.SESSION_TOKEN)
        self.nso._session_token = token

    def generate_gtoken(self) -> None:
        """Generates a new gtoken from the internal NSO class and adds it to the
        manager. This is a convenience method that will also set the gtoken in
        the NSO class. This function requires a session token to already be
        set, and will raise a ValueError if it is not.

        Raises:
            ValueError: If the session token has not been set.
            NintendoException: In the unlikely event that the gtoken has made it
                through the NSO class but the user info was not set, this error
                provides a safety net to identify that there was a problem with
                the gtoken generation. Should not be raised in normal use.
        """
        if TOKENS.SESSION_TOKEN not in self._tokens:
            raise ValueError(
                "Session token must be set before generating a gtoken."
            )
        gtoken = self.nso.get_gtoken(self.nso.session_token)
        self.add_token(gtoken, TOKENS.GTOKEN)
        try:
            user_info = cast(dict[str, str], self.nso._user_info)
            country = user_info["country"]
            language = user_info["language"]
            self._data["country"] = country
            self._data["language"] = language
        except (KeyError, TypeError):
            raise NintendoException(
                "Unable to get user info. Gtoken may be invalid."
            )

    @retry(times=1, exceptions=SplatNetException)
    def generate_bullet_token(self) -> None:
        """Generates a new bullet token from the internal NSO class and adds it
        to the manager. This is a convenience method that will also generate a
        ``gtoken`` if one has not already been generated, or if the user info
        has not been set. This function requires a session token to already be
        set, and will raise a ValueError if it is not. If the bullet token is
        unable to be generated, a SplatNetException will be raised.

        Raises:
            ValueError: If the session token has not been set.
            SplatNetException: If the bullet token was unable to be generated.
        """
        if TOKENS.SESSION_TOKEN not in self._tokens:
            raise ValueError(
                "Session token must be set before generating a bullet token."
            )
        if (TOKENS.GTOKEN not in self._tokens) or (self.nso._user_info is None):
            self.generate_gtoken()
        bullet_token = self.nso.get_bullet_token(
            cast(str, self.nso._gtoken), cast(dict, self.nso._user_info)
        )
        self.add_token(bullet_token, TOKENS.BULLET_TOKEN)
        bullet = self.get(TOKENS.BULLET_TOKEN, full_token=True)
        if (bullet is not None) and not bullet.is_valid:
            raise SplatNetException(
                "Bullet token was unable to be generated. This is likely due "
                "to SplatNet 3 being down. Please try again later."
            )

    def generate_all_tokens(self) -> None:
        """Generates all tokens that can be generated from the manager. This
        is a convenience method that will call the ``generate_gtoken`` and
        ``generate_bullet_token`` methods in order. This function requires a
        session token to already be set, and will raise the errors that those
        methods raise if a session token is not set.
        """
        self.generate_gtoken()
        self.generate_bullet_token()

    @classmethod
    def from_session_token(cls, session_token: str) -> "TokenManager":
        """Creates a token manager from a session token.

        Given a session token, this method will create a token manager and add
        a session token to it. Additionally, it will flag the origin of the
        session token as "session_token" so that it can be saved to a config
        file later.

        Args:
            session_token (str): The session token to use to instantiate the
                ``TokenManager``.

        Returns:
            TokenManager: The token manager with the session token added.
        """
        manager = cls()
        manager.add_session_token(session_token)
        manager.flag_origin("session_token")
        return manager

    @classmethod
    def load(cls) -> "TokenManager":
        """Loads tokens from a config file or environment variables.

        Checks for appropriate tokens in the following order:
            1. .splatnet3_scraper file
            2. Environment variables
            3. tokens.ini file

        Raises:
            ValueError: If no tokens are found.

        Returns:
            TokenManager: The token manager with the tokens loaded.
        """
        if os.path.exists(".splatnet3_scraper"):
            return cls.from_config_file(".splatnet3_scraper")
        elif any([os.environ.get(var) for var in ENV_VAR_NAMES.values()]):
            return cls.from_env()
        elif os.path.exists("tokens.ini"):
            return cls.from_config_file("tokens.ini")
        else:
            raise ValueError(
                "No tokens found. Please create a .splatnet3_scraper file, set "
                "environment variables, or create a tokens.ini file."
            )

    @classmethod
    def from_config_file(cls, path: str) -> "TokenManager":
        """Loads tokens from a config file.

        Given a path to a config file, this method will create a token manager
        and add the tokens found in the config file to it. Additionally, it
        will flag the origin of the tokens as "config_file" so that they can be
        saved to a config file later. The config file must be in the format of
        the standard ``configparser`` library, and must have a "tokens" section.
        The "tokens" section must have a "session_token" option, and may have
        "gtoken" and "bullet_token" options. An example config file is shown
        below:

        >>> [tokens]
        ... session_token = SESSION_TOKEN
        ... gtoken = GTOKEN
        ... bullet_token = BULLET_TOKEN
        ...
        ... [data]
        ... country = US
        ... language = en-US
        ...
        ... [options]
        ... user_agent = USER_AGENT

        Tests the tokens before returning the token manager using the
        ``test_tokens`` method.

        Args:
            path (str): The path to the config file.

        Raises:
            ValueError: If the config file does not have a ``tokens`` section.

        Returns:
            TokenManager: A newly created token manager with the tokens loaded
                and the origin of the token manager set to "config_file".
        """
        config = configparser.ConfigParser()
        config.read(path)
        nso = NSO.new_instance()
        tokenmanager = cls(nso)
        tokenmanager.flag_origin("config_file", path)

        if not config.has_section("tokens"):
            raise ValueError("Config file does not have a 'tokens' section.")
        for option in config.options("tokens"):
            token = config.get("tokens", option)
            if option == TOKENS.SESSION_TOKEN:
                nso._session_token = token
            elif option == TOKENS.GTOKEN:
                nso._gtoken = token
            tokenmanager.add_token(token, option)

        if not config.has_section("data"):
            tokenmanager.generate_all_tokens()
            return tokenmanager
        for option in config.options("data"):
            tokenmanager._data[option] = config.get("data", option)
        tokenmanager.test_tokens()
        return tokenmanager

    @classmethod
    def from_text_file(cls, path: str) -> "TokenManager":
        """Loads tokens from a text file, particularly s3s config files. Not
        recommended for use, but here for compatibility with s3s config files.

        Requires the text file to be a JSON file that contains a "session_token"
        key. The value of the "session_token" key will be used as the session
        token. If the text file also contains an "acc_loc" key, the value of
        the "acc_loc" key will be used to set the language and country of the
        data in the token manager.

        Tests the tokens before returning the token manager using the
        ``test_tokens`` method.

        Args:
            path (str): The path to the text file.

        Raises:
            ValueError: If the session token is not found in the JSON file.

        Returns:
            TokenManager: The token manager with the tokens loaded.
        """
        token_manager = cls()
        with open(path, "r") as f:
            data = json.load(f)

        if "session_token" not in data:
            raise ValueError("Session token not found in text file.")
        token_manager.add_session_token(data["session_token"])
        token_manager.flag_origin("text_file", path)
        if "acc_loc" in data:
            language, country = data["acc_loc"].split("|")
            token_manager._data["language"] = language
            token_manager._data["country"] = country

        if "gtoken" in data:
            token_manager.add_token(data["gtoken"], TOKENS.GTOKEN)
        if "bullettoken" in data:
            token_manager.add_token(data["bullettoken"], TOKENS.BULLET_TOKEN)
        token_manager.test_tokens()
        return token_manager

    @classmethod
    def from_env(
        cls, env_manager: EnvironmentVariablesManager | None = None
    ) -> "TokenManager":
        """Loads tokens from environment variables.

        This method will create a token manager and add the tokens found in the
        environment variables to it. The environment variables that are
        supported are:

        - SN3S_SESSION_TOKEN
        - SN3S_GTOKEN
        - SN3S_BULLET_TOKEN

        The session token environment variable is required, and if it is not
        set, a ValueError will be raised. The other environment variables are
        optional and will be generated if they are not set.

        Tests the tokens before returning the token manager using the
        ``test_tokens`` method.

        Args:
            env_manager (EnvironmentVariablesManager): The environment variables
                manager to use. If not provided, a new one will be created.

        Raises:
            ValueError: If the session token environment variable is not set.

        Returns:
            TokenManager: The token manager with the tokens loaded.
        """
        nso = NSO.new_instance()
        tokenmanager = cls(nso, env_manager=env_manager)
        tokens = tokenmanager.env_manager.get_all()
        for token, value in tokens.items():
            if token == TOKENS.SESSION_TOKEN:
                if value is None:
                    raise ValueError(
                        "Session token environment variable not set."
                    )
                nso._session_token = value
            elif value is None:
                continue
            elif token == TOKENS.GTOKEN:
                nso._gtoken = value
            tokenmanager.add_token(value, token)
        tokenmanager.flag_origin("env")
        tokenmanager.test_tokens()
        return tokenmanager

    def save(self, path: str | None = None) -> None:
        """Saves the tokens to a config file.

        Uses the ``configparser`` module to save the tokens to a config file.
        The config file will be saved to the path specified by the `path`
        argument. If the `path` argument is not specified, the config file
        will be saved to the current working directory with the name
        ``.splatnet3_scraper``. Does not consider the origin of the tokens, as
        it is assumed that this method will only be called when the user wants
        to save the tokens to disk.

        Args:
            path (str): The path to the config file.
        """
        config = configparser.ConfigParser()
        out_tokens = {}
        for token_name, token in self._tokens.items():
            out_tokens[token_name] = token.token
        config["tokens"] = out_tokens
        config["data"] = self._data
        config["metadata"] = {
            "version": __version__,
            "class": self.__class__.__name__,
        }
        if path is None:
            path = ".splatnet3_scraper"
        with open(path, "w") as configfile:
            config.write(configfile)

    def token_is_valid(self, token_type: str) -> bool:
        """Given a token type, checks if the token is valid.

        This method will check if the token is valid by checking if the token
        is expired. This method will not regenerate the token if it is invalid.
        As mentioned in the ``Token.is_valid`` property, this method will not
        check if the token is valid by making a request to the API, but rather
        checks if the token is likely to be expired by comparing the known
        expiration time of the token to the current time. If the token type is
        not found, this method will return False.

        Args:
            token_type (str): The type of token to check.

        Returns:
            bool: True if the token is valid, False otherwise.
        """
        try:
            token = self.get(token_type, full_token=True)
        except ValueError:
            return False
        return token.is_valid

    def test_tokens(self, user_agent: str | None = None) -> None:
        """Tests the tokens.

        Checks using the ``token_is_valid`` method to see whether the tokens
        need to be regenerated. If so, the tokens will be regenerated with the
        appropriate ``generate_*`` method. Once the tokens are regenerated, it
        will make a request to the API to check if the tokens are valid. If the
        request fails, the ``generate_all_tokens`` method will be called. If the
        requests succeeds, nothing will happen. Once this method is called, the
        TokenManager can be assumed to have valid tokens and can be used to make
        requests to the API.

        Args:
            user_agent (str): The user agent to use when making the request to
                the API. If not specified, the default user agent will be used.

        Raises:
            ValueError: If the session token is not set.
        """
        if self.get(TOKENS.SESSION_TOKEN) is None:
            raise ValueError("Session Token is not set.")

        if self.token_is_valid(TOKENS.GTOKEN) is False:
            self.generate_gtoken()

        if self.token_is_valid(TOKENS.BULLET_TOKEN) is False:
            self.generate_bullet_token()

        header = queries.query_header(
            self.get(TOKENS.BULLET_TOKEN), self._data["language"], user_agent
        )

        response = requests.post(
            GRAPH_QL_REFERENCE_URL,
            data=queries.query_body("HomeQuery"),
            headers=header,
            cookies={"_gtoken": cast(str, self.get(TOKENS.GTOKEN))},
        )
        if response.status_code != 200:
            self.generate_all_tokens()



import logging
import re
import time
from functools import lru_cache, wraps
from typing import Any, Callable, ParamSpec, Type, TypeAlias, TypeVar, cast

import requests

from splatnet3_scraper.constants import (
    GRAPH_QL_REFERENCE_URL,
    HASHES_FALLBACK,
    WEB_VIEW_VERSION_FALLBACK,
)

T = TypeVar("T")
P = ParamSpec("P")

PathType: TypeAlias = str | int | tuple[str | int, ...]

json_splitter_re = re.compile(r"[\;\.]")


def retry(
    times: int,
    exceptions: tuple[Type[Exception], ...] | Type[Exception] = Exception,
    call_on_fail: Callable[[], None] | None = None,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """Decorator that retries a function a specified number of times if it
    raises a specific exception or tuple of exceptions.

    Args:
        times (int): Max number of times to retry the function before raising
            the exception.
        exceptions (tuple[Type[Exception], ...] | Type[Exception]): Exception
            or tuple of exceptions to catch. Defaults to Exception.
        call_on_fail (Callable[[], None] | None): Function to call if the
            function fails. If None, nothing will be called. Defaults to None.

    Returns:
        Callable[[Callable[P, T]], Callable[P, T]]: The decorated function,
            which will retry the function if it raises an exception.
    """

    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            for i in range(times):
                try:
                    return func(*args, **kwargs)
                except exceptions:
                    logging.warning(
                        "%s failed on attempt %d of %d, retrying.",
                        func.__name__,
                        i + 1,
                        times + 1,
                    )
                    if call_on_fail is not None:
                        logging.debug("Calling %s...", call_on_fail.__name__)
                        call_on_fail()

            return func(*args, **kwargs)

        return wrapper

    return decorator


def linearize_json(
    json_data: dict[str, Any]
) -> tuple[tuple[str, ...], list[Any]]:
    """Linearizes a JSON object. Given a JSON object, this function will
    return a tuple of keys and values. Turns a JSON object into a table.

    Given a JSON object, this function will return a tuple of keys and values.
    The keys will be a tuple of strings, and the values will be a list of
    values. The keys will be the path to the value in the JSON object, and the
    values will be the values of the JSON object. For example, given the
    following JSON object:

    >>> json_data = {
    ...     "a": 1,
    ...     "b": {
    ...         "c": 2,
    ...         "d": 3,
    ...     },
    ...     "e": [4, 5],
    ... }

    The function will return the following tuple:

    >>> linearize_json(json_data)
    ... (
    ...     ("a", "b.c", "b.d", "e;0", "e;1"),
    ...     [1, 2, 3, 4, 5],
    ... )

    Args:
        json_data (dict[str, Any]): The JSON object to linearize.

    Returns:
        tuple[str, ...]: The keys of the JSON object. The keys are in the
            format of "key1.key2;index1.key3" where the semicolon indicates a
            list and the period indicates a nested object.
        list[Any]: The values of the JSON object. The values are in the same
            order as the keys.
    """
    keys = []
    values = []

    for key, value in json_data.items():
        if isinstance(value, dict):
            sub_keys, sub_values = linearize_json(value)
            keys.extend([(key + "." + sub_key) for sub_key in sub_keys])
            values.extend(sub_values)
        elif isinstance(value, list):
            for i, item in enumerate(value):
                if isinstance(item, dict):
                    sub_keys, sub_values = linearize_json(item)
                    keys.extend(
                        [
                            (key + ";" + str(i) + "." + sub_key)
                            for sub_key in sub_keys
                        ]
                    )
                    values.extend(sub_values)
                else:
                    keys.append(key + ";" + str(i))
                    values.append(item)
        else:
            keys.append(key)
            values.append(value)

    # Turn the keys into an immutable tuple so it can be hashed
    out_keys = tuple(keys)
    return out_keys, values


def delinearize_json(
    keys: list[str] | tuple[str, ...], values: list[Any]
) -> dict[str, Any]:
    """Delinearizes a JSON object. Given a list of keys and a list of values,
    this function will return a JSON object. Turns a table into a JSON object.

    Given a list of keys and a list of values, this function will return a
    JSON object. The keys are expected to be in the format of
    "key1.key2;index1.key3" where the semicolon indicates a list and the
    period indicates a nested object. The values are expected to be in the
    same order as the keys.

    Args:
        keys (list[str]): The keys of the JSON object. The keys are expected
            to be in the format of "key1.key2;index1.key3" where the semicolon
            indicates a list and the period indicates a nested object.
        values (list[Any]): The values of the JSON object. The values are
            expected to be in the same order as the keys.

    Returns:
        dict[str, Any]: The JSON object created from the keys and values.
    """
    json_data = {}

    # Sort the keys by depth
    depths = [len(json_splitter_re.split(key)) for key in keys]
    kvd = list(zip(keys, values, depths))
    kvd = sorted(kvd, key=lambda x: (x[2], x[0]))
    keys, values, _ = list(zip(*kvd))

    # Delinearize
    for key, value in zip(keys, values):
        # If the key is split by a period, it's a nested object. If it's split
        # by a semicolon, it's a list. Check which one is first.
        subkeys = json_splitter_re.split(key)
        splitters = json_splitter_re.findall(key)
        if len(subkeys) == 1:
            json_data[key] = value
            continue
        # If the key is split by a semicolon, turn the next key value into an
        # integer
        for i, splitter in enumerate(splitters):
            if splitter == ";":
                subkeys[i + 1] = int(subkeys[i + 1])

        current = json_data
        for i, splitter in enumerate(splitters):
            # If the key already exists, move on to the next key
            if isinstance(current, (list, dict)):
                chosen_subkey = subkeys[i]
                if isinstance(current, list):
                    condition = len(current) > chosen_subkey
                elif isinstance(current, dict):
                    condition = chosen_subkey in current

                if condition:
                    next_obj = current[chosen_subkey]
                    # Next object might be None as an artifact of header merging
                    if next_obj is not None:
                        current = next_obj
                        continue
            new_obj: dict | list = {} if (splitter == ".") else []

            # Append or assign the new object to the current object depending
            # on type.
            if isinstance(current, list):
                current.append(new_obj)
            elif isinstance(current, dict):
                current[subkeys[i]] = new_obj
            # Mypy doesn't like that this is dynamically typed, so we have to
            # ignore it. This is python, not C.
            current = new_obj  # type: ignore
        if isinstance(current, list):
            current.append(value)
        else:
            current[subkeys[-1]] = value

    return json_data


def enumerate_all_paths(data: dict | list | Any) -> list[tuple[str | int, ...]]:
    """Recursively enumerates all paths in the given data.

    Given a dictionary or list, returns a list of all paths in the data. For
    example, given the following data:

    >>> data = {
    ...     "a": 1,
    ...     "b": {
    ...         "c": 2,
    ...         "d": [3, 4],
    ...     },
    ... }

    The following paths would be returned:

    >>> enumerate_all_paths(data)
    ... [
    ...     ("a",),
    ...     ("b",),
    ...     ("b", "c"),
    ...     ("b", "d"),
    ...     ("b", "d", 0),
    ...     ("b", "d", 1),
    ... ]

    If the given data is not a dictionary or list, an empty list is returned.

    Args:
        data (dict | list | Any): The data to enumerate. If the data is not a
            dictionary or list, an empty list is returned. Otherwise, the
            function recurses into the data.

    Returns:
        list[tuple[str | int, ...]]: A list of all paths in the given data.
    """
    paths: list[tuple[str | int, ...]] = []

    if not isinstance(data, (dict, list)):
        return paths

    members = data.items() if isinstance(data, dict) else enumerate(data)

    for key, value in members:
        key_tuple = (cast(str | int, key),)
        paths.append(key_tuple)
        subpaths = enumerate_all_paths(value)
        for subpath in subpaths:
            paths.append(key_tuple + subpath)

    return paths


def match_partial_path(
    data: dict[str, Any] | list, partial_path: PathType | list[PathType]
) -> list[tuple[str | int, ...]]:
    """Returns a list of all paths in the given data that match the given
    partial path. For example, if partial_path is ``(0, "key1")``, this will
    return all paths in the data that match ``...[0]["key1"]``. If fed a list
    of partial paths, this will return all paths that match any of the partial
    paths. Do not confuse tuples with lists, as they are treated differently.

    The ":" string can be added to the partial path input to match all list
    indices that match it. This is useful for JSONs with a repeating structure.

    The match_partial_path algorithm searches for all paths in a dictionary that
    match the given partial path. The ``partial_path`` can be a string, integer,
    a special ":" string, or a tuple of strings/integers that represents the
    path to an item in the dictionary. For example, the path
    ``("key1", "key2", 2)`` corresponds to ``...["key1"]["key2"][2]`` in the
    dictionary. The algorithm returns a list of all paths that match the
    ``partial_path``. Each path in the result is represented as a tuple of
    strings and integers where strings correspond to dictionary keys and
    integers correspond to list indices.

    For instance, if ``data`` is:

    >>> data = {
    ...     "key1": {
    ...         "key2": [
    ...             {"key3": 1},
    ...             {"key3": 2},
    ...         ]
    ...     },
    ...     "key4": {
    ...         "key5": [
    ...             {"key3": 3},
    ...             {"key3": 4},
    ...         ]
    ...     }
    ... }

    Then ``match_partial_path(data, (0, "key3"))`` will return:

    >>> [
    ...     ("key1", "key2", 0, "key3"),
    ...     ("key4", "key5", 0, "key3"),
    ... ]

    If ``match_partial_path(data, "key3")`` is called, the result will be:

    >>> [
    ...     ("key1", "key2", 0, "key3"),
    ...     ("key1", "key2", 1, "key3"),
    ...     ("key4", "key5", 0, "key3"),
    ...     ("key4", "key5", 1, "key3"),
    ... ]

    If ``match_partial_path(data, [(0, "key3"), "key2"])`` is called, the result
    will be:

    >>> [
    ...     ("key1", "key2", 0, "key3"),
    ...     ("key4", "key5", 0, "key3"),
    ...     ("key1", "key2"),
    ... ]

    If ``match_partial_path(data, ("key1", "key2", ":", "key3"))`` is called,
    the result will be:

    >>> [
    ...     ("key1", "key2", 0, "key3"),
    ...     ("key1", "key2", 1, "key3"),
    ... ]

    Args:
        data (dict[str, Any] | list): The dictionary or list to search.
        partial_path (PathType | list[PathType]): The partial path to match. If
            the partial path is a tuple, this function will treat it as a path.
            For example, a ``partial_path`` argument of ``(0, "key1")`` will be
            treated  as ``...[0]["key1"]``. If the partial path is a string,
            this function will treat it as a key in a dictionary. Integers will
            be treated as indices in a list. If the partial path is a list, this
            function will return all paths that match any of the partial paths.
            The special ":" string can be used to match all list indices that
            match the rest of the partial path. This is useful to get all paths
            in a list that match a given partial path.

    Returns:
        list[tuple[str | int, ...]]: A list of all paths that match the given
            partial path. Each path is represented as a tuple of strings and
            integers, where strings correspond  to dictionary keys and integers
            correspond to list indices.
    """
    if isinstance(partial_path, list):
        short_circuit: list[tuple[str | int, ...]] = []
        for path in partial_path:
            short_circuit.extend(match_partial_path(data, path))
        return short_circuit

    if isinstance(partial_path, (str, int)):
        partial_path = (partial_path,)
    partial_path = tuple(partial_path)

    # Check if the partial path contains the ":" character, which is used to
    # denote to allow all indices in a list. If it does, then we need to
    # enumerate all paths in the data and filter out the ones that match the
    # partial path.
    branching_points = len([i for i, x in enumerate(partial_path) if x == ":"])

    # If the partial path is empty, return an empty list
    paths = enumerate_all_paths(data)
    out: list[tuple[str | int, ...]] = []

    if branching_points == 0:
        for path in paths:
            if path[-len(partial_path) :] == partial_path:
                out.append(path)

        return out

    # If the partial path contains the ":" character, then we need to
    # enumerate all paths in the data and filter out the ones that match the
    # partial path.
    for path in paths:
        add_path = True
        truncated_path = path[-len(partial_path) :]
        if len(truncated_path) < len(partial_path):
            continue
        for idx, value in enumerate(partial_path):
            if value == truncated_path[idx]:
                continue
            elif value == ":":
                if isinstance(truncated_path[idx], int):
                    continue

            add_path = False
            break
        if add_path:
            out.append(path)

    return out


@lru_cache()
def get_hash_data(
    url: str | None = None, ttl_hash: int | None = None
) -> tuple[dict, str]:
    """Gets the hash data for the GraphQL queries with a time-to-live (TTL)
    cache.

    Uses requests to get the `imink` GraphQL query hash map JSON file and
    parses it to get the hashes for the queries. The initial request
    response contains two keys: ``hash_map`` and ``version``. Both of these
    are returned as a tuple, with the first element being the ``hash_map``
    and the second element being the ``version``.

    Args:
        url (str | None): The URL to get the hash data from. If None, the
            default URL will be used, from `imink`. This can be found in the
            ``GRAPH_QL_REFERENCE_URL`` variable. Defaults to None.
        ttl_hash (int | None): The hash to use for the TTL cache. This is used
            to determine if the cache should be used or not. The hash is
            calculated by dividing the current time by the expiry time in
            seconds and rounding to the nearest integer. If None, the default
            expiry time of 15 minutes will be used. Note that this method of TTL
            caching does not guarantee that the cache will be invalidated after
            the expiry time, but it is a good enough approximation. Defaults to
            None.

    Returns:
        dict[str, str]: The hash map for the GraphQL queries.
        str: The version of the hash map.
    """
    del ttl_hash

    request_url = url or GRAPH_QL_REFERENCE_URL
    response = requests.get(request_url).json()
    return response["graphql"]["hash_map"], response["version"]


def get_ttl_hash(expiry_time_seconds: float = 15 * 60) -> int:
    return round(time.time() / expiry_time_seconds)


def get_splatnet_hashes(url: str | None = None) -> dict[str, str]:
    """Gets the hashes for the GraphQL queries.

    Uses requests to get the `imink` GraphQL query hash map JSON file and
    parses it to get the hashes for the queries. The initial request
    response contains two keys: ``hash_map`` and ``version``. The
    ``hash_map`` key is the dictionary that contains the hashes for the
    queries and is what is returned by this method. The `version` key is the
    version of the hashes and is used to check if the hashes are up to date,
    it is used elsewhere in this package, but is not used here.

    Args:
        url (str | None): The URL to get the hash data from. If None, the
            default URL will be used, from `imink`. This can be found in the
            ``GRAPH_QL_REFERENCE_URL`` constant.

    Returns:
        dict[str, str]: The hashes for the GraphQL queries. The keys are
            the names of the queries and the values are the most up to date
            hashes for the queries.
    """
    try:
        hash_data, _ = get_hash_data(url, get_ttl_hash())
    except Exception as e:
        logging.warning(f"Failed to get hash data: {e}")
        logging.warning("Using fallback")
        return HASHES_FALLBACK
    return hash_data


def get_splatnet_version(url: str | None = None) -> str:
    """Gets the version of the GraphQL queries.

    Uses requests to get the `imink` GraphQL query hash map JSON file and
    parses it to get the version of the queries. The initial request
    response contains two keys: ``hash_map`` and ``version``. The
    ``version`` key is the version of the hashes and is what is returned by
    this method. The `hash_map` key is the dictionary that contains the
    hashes for the queries and is used elsewhere in this package, but is not
    used here.

    Args:
        url (str | None): The URL to get the hash data from. If None, the
            default URL will be used, from `imink`. This can be found in the
            ``GRAPH_QL_REFERENCE_URL`` constant.

    Returns:
        str: The version of the GraphQL queries.
    """
    try:
        _, version = get_hash_data(url, get_ttl_hash())
    except Exception as e:
        logging.warning(f"Failed to get hash data: {e}")
        logging.warning("Using fallback")
        return WEB_VIEW_VERSION_FALLBACK
    return version



import base64
import hashlib
import logging
import os
import re
from typing import Callable, Literal, TypeAlias, cast

import requests

from splatnet3_scraper import __version__
from splatnet3_scraper.auth.exceptions import (
    FTokenException,
    NintendoException,
    SplatNetException,
)
from splatnet3_scraper.constants import (
    APP_VERSION_FALLBACK,
    DEFAULT_USER_AGENT,
    IMINK_URL,
    IOS_APP_URL,
    SPLATNET_URL,
    WEB_VIEW_VERSION_FALLBACK,
)
from splatnet3_scraper.utils import get_splatnet_version, retry

version_re = re.compile(
    r"(?<=whats\-new\_\_latest\_\_version\"\>Version)\s+\d+\.\d+\.\d+"
)

FToken_Gen: TypeAlias = Callable[
    [str, str, Literal[1] | Literal[2], str, str | None],
    tuple[str, str, str],
]


class NSO:
    """The NSO class contains all the logic to proceed through the login flow.
    This class also holds various properties that are used to make requests to
    the Nintendo Switch Online API. Login flow is roughly as follows, assuming
    the user has never generated a session token before:

        1.  Initialize a requests session and store it.
        2.  Generate a random state and S256 code challenge that will be used
            to obtain the "session_token". Store them for later use.
        3.  Generate a login URL using the state and code challenge that the
            user will open in their browser. The user will then copy a link and
            feed it back to the program.
        4.  Parse the URI to obtain the session token code, then use it
            alongside the code challenge to obtain the ``session_token``. Store
            it for later use. The session token is valid for 2 years and can be
            revoked by the user.
        5.  Use the session token to obtain a user access response from
            Nintendo. This response will contain an ``id_token`` and a user
            access token. Store both for later user.
        6.  Use the user access token to obtain the user information. This is
            required to obtain the first ``f_token``, for the Nintendo Switch
            Online API. This is a message authentication code generated from the
            timestamp and id token in an obscure manner and will be necessary
            to generate the API access token.
        7.  Use the Nintendo-obtained ``id_token`` to obtain the first
            ``f_token``, for the Nintendo Switch Online API. The ``f_token``
            will also return a timestamp and a request ID.
        8.  Use the first ``f_token``, the request ID, the timestamp, the user
            information, the ``id_token`` to obtain the
            ``web_service_access_token``.
        9.  Use the ``web_service_access_token`` to obtain the second
            ``f_token``. This is a message authentication code generated from
            the timestamp and id token in an obscure manner and will be
            necessary to generate the ``gtoken``. This step is distinct from
            step 7, make sure not to use the same ``f_token`` twice.
        10. Use the second ``f_token``, the request ID, the timestamp, and the
            ``web_service_access_token`` from step 9, and the id
            ``4834290508791808`` to obtain the ``gtoken``.
        11. Use the ``gtoken`` and user information to obtain a
            ``bullet_token``. This token is valid for 2 hours.

    Once the login flow is complete, the NSO class contains all the necessary
    values to regenerate the tokens. To minimize the number of objects to track,
    the NSO class only stores the tokens in memory. The TokenManager class
    handles the persistence of the tokens to disk.
    """

    def __init__(self, session: requests.Session) -> None:
        """Initializes the NSO class. The NSO class contains all the logic and
        holds all the necessary values to proceed through the login flow.

        The ``__init__`` method sets the following internal variables:

            -   ``session``: The requests session object that is used to make
                requests to the Nintendo Switch Online API. This is the only
                variable that is passed in to the ``__init__`` method.
            -   ``_state``: The internal state variable that is used to obtain
                the session token. This is a random string that is generated
                when the ``NSO.state`` property is first accessed, and is used
                to generate the login URL.
            -   ``_verifier``: The internal verifier variable that is used to
                obtain the session token. This is a random string that is
                generated when the ``NSO.verifier`` property is first accessed,
                and is used to generate the login URL. The verifier is also
                used to solve the code challenge, verifying to Nintendo that
                the user is who they say they are.
            -   ``_version``: The internal version variable that is used to
                verify the NSO app version. This is a string that is generated
                when the ``NSO.version`` property is first accessed, and is
                used to obtain the session token.
            -   ``_web_view_version``: The internal web view version variable
                that is used to verify the NSO app version. This is a string
                that is generated when the ``NSO.web_view_version`` property is
                first accessed, and is used to obtain the session token.
            -   ``_session_token``: A stored session token. This is required to
                generate all other tokens and information. This is obtained
                during the login flow and is valid for 2 years.
            -   ``_user_access_token``: A stored user access token. This is a
                token that is obtained during the login flow and is used to
                obtain the user information.
            -   ``_id_token``: A stored id token. This is a token that is
                obtained during the login flow and is used to obtain the gtoken.
            -   ``_gtoken``: A stored gtoken. This is a token that is obtained
                during the login flow and is used to obtain the bullet token. It
                is valid for 6 hours and 30 minutes.
            -   ``_user_info``: A stored user information dictionary. This is
                obtained during the login flow and is used to obtain the "f"
                token.
            -   ``_f_token_function``: A stored function that is used to obtain
                the "f" token. This function is set to the ``get_ftoken``
                method by default, and can be set with the
                ``set_new_f_token_function`` method to use a user-defined
                function if desired.


        Args:
            session (requests.Session): The NSO class uses a requests session
                to make requests to the Nintendo Switch Online API. The session
                object is passed in to the NSO class so that the same session
                object can be used to make requests to the SplatNet API.
        """
        self.session = session
        self._state: bytes | None = None
        self._verifier: bytes | None = None
        self._version: str | None = None
        self._web_view_version: str | None = None
        self._session_token: str | None = None
        self._user_access_token: str | None = None
        self._id_token: str | None = None
        self._nintendo_account_id: str | None = None
        self._coral_user_id: str | None = None
        self._gtoken: str | None = None
        self._user_info: dict[str, str] | None = None
        self._f_token_function: FToken_Gen = self.get_ftoken

    @staticmethod
    def new_instance() -> "NSO":
        """Creates a new instance of the NSO class with a new requests session.

        This is the recommended way to create a new instance of the NSO class,
        as it ensures that the session is a fresh session, however it is not
        absolutely necessary to instantiate a new NSO class using this method.
        Passing in a new session to the ``__init__`` method is perfectly fine.

        Returns:
            NSO: A new instance of the NSO class.
        """
        session = requests.Session()
        return NSO(session=session)

    @property
    def version(self) -> str:
        """Returns the current version of the NSO app. Necessary to get the
        session token. If the version has not been obtained yet, it will be
        obtained and stored. If the version cannot be obtained, a fallback
        version will be used.

        Returns:
            str: The current version of the NSO app.
        """
        if self._version is None:
            self._version = self.get_version()
        return self._version

    @retry(times=2, exceptions=ValueError)
    def get_version(self) -> str:
        """Fetches the current version of the Nintendo Switch Online app from
        the iOS app store. This is necessary to access the API. This method
        retries twice if the version cannot be obtained, in case the iOS App
        Store site is down or slow. If the version cannot be obtained after
        three attempts, a fallback version defined in the ``constants.py`` file
        will be used.

        Returns:
            str: The current version of the NSO app.
        """
        # TODO: Replace the iOS app store method with a method that does not
        # require scraping a website with scraping protection.
        response = self.session.get(IOS_APP_URL)
        version = version_re.search(response.text)
        if version is None:
            logging.warning(
                "Failed to get version from app store, using fallback"
            )
            return APP_VERSION_FALLBACK
        return version.group(0).strip()

    @property
    def state(self) -> bytes:
        """Returns a base64url encoded random 36 byte string for the auth state.
        This is used to generate the login URL. To align with node.js's crypto
        module, padding is removed. If the state has not been obtained yet, it
        will be obtained and stored by creating a random 36 byte string and then
        base64url encoding it.

        Returns:
            bytes: The auth state, without padding. This is used within the
                login flow to verify that the user is who they say they are.
        """
        if self._state is None:
            self._state = self.generate_new_state()
        return self._state

    def generate_new_state(self) -> bytes:
        """Generates a new state.

        Returns:
            bytes: The auth state, without padding. A random 36 byte string
                that is base64url encoded.
        """
        return base64.urlsafe_b64encode(os.urandom(36))

    @property
    def verifier(self) -> bytes:
        """Returns a base64url encoded random 32 byte string for the code
        verifier. This is used to generate the S256 code challenge. To align
        with node.js's crypto module, padding is removed. If the verifier has
        not been obtained yet, it will be obtained and stored by creating a
        random 32 byte string and then base64url encoding it.

        Returns:
            bytes: The code verifier, without padding. This is used within the
                login flow to verify that the user is who they say they are.
        """
        if self._verifier is None:
            self._verifier = self.generate_new_verifier()
        return self._verifier

    def generate_new_verifier(self) -> bytes:
        """Generates a new code verifier, which is a random 32 byte string
        that is base64url encoded and with padding removed. This is used to
        generate the S256 code challenge.

        Returns:
            bytes: The code verifier, without padding.
        """
        verifier_ = base64.urlsafe_b64encode(os.urandom(32))
        return verifier_.replace(b"=", b"")

    @property
    def session_token(self) -> str:
        """Returns the session token. This cannot be generated and must be set
        by the user. If the session token has not been set, a ValueError will be
        raised.

        Raises:
            ValueError: The session token has not been set.

        Returns:
            str: The session token.
        """
        if self._session_token is None:
            raise ValueError("Session token is not set.")
        return self._session_token

    def generate_login_url(self, user_agent: str | None = None) -> str:
        """Generates the login URL that can be used to obtain the session token.

        This method will generate the login URL that is used to obtain the
        session token. For full details on how to obtain a session token, see
        the ``Obtaining a session token`` section of the documentation.

        Args:
            user_agent (str): The user agent to use. Defaults to the default
                user agent found in ``constants.py``.

        Returns:
            str: The login URL that can be used to obtain the session token.
        """
        # https://dev.to/mathewthe2/intro-to-nintendo-switch-rest-api-2cm7
        hash_ = hashlib.sha256()
        hash_.update(self.verifier)
        challenge = base64.urlsafe_b64encode(hash_.digest()).replace(b"=", b"")

        header = {
            "Host": "accounts.nintendo.com",
            "Connection": "keep-alive",
            "Cache-Control": "max-age=0",
            "Upgrade-Insecure-Requests": "1",
            "User-Agent": user_agent
            if user_agent is not None
            else DEFAULT_USER_AGENT,
            "Accept": (
                "text/html,"
                + "application/xhtml+xml,"
                + "application/xml;q=0.9,"
                + "image/webp,"
                + "image/apng,"
                + "*/*;q=0.8n"
            ),
            "DNT": "1",
            "Accept-Encoding": "gzip,deflate,br",
        }
        params = {
            "state": self.state,
            "redirect_uri": "npf71b963c1b7b6d119://auth",
            "client_id": "71b963c1b7b6d119",
            "scope": ("openid user user.birthday user.mii user.screenName"),
            "response_type": "session_token_code",
            "session_token_code_challenge": challenge,
            "session_token_code_challenge_method": "S256",
            "theme": "login_form",
        }
        login_url = "https://accounts.nintendo.com/connect/1.0.0/authorize"
        response = self.session.get(
            login_url, headers=header, params=params  # type: ignore
        )
        return response.url

    def parse_npf_uri(self, uri: str) -> str:
        """Parses the uri returned by the Nintendo login page and extracts the
        session token code. This is used to pass the challenge and verify that
        the user is who they say they are.

        Args:
            uri (str): The uri returned by the Nintendo login page.

        Returns:
            str: The session token code. This is *NOT* the session token, but is
                used to obtain the session token.
        """
        return uri.split("&")[1][len("session_token_code=") :]

    def get_session_token(self, session_token_code: str) -> str:
        """Obtains the session token from the session token code.

        This method will obtain the session token from the session token code.
        It will use provide the session token code to Nintendo's servers
        alongside the session token code verifier. Nintendo will then verify
        that the user is who they say they are and return the session token.

        Args:
            session_token_code (str): The session token code. This is obtained
                by parsing the uri obtained at the URL returned by
                ``generate_login_url``.

        Returns:
            str: The session token. DO NOT SHARE THIS TOKEN WITH ANYONE.
        """
        header = {
            "User-Agent": f"OnlineLounge/{self.version} NASDKAPI Android",
            "Accept-Language": "en-US",
            "Accept": "application/json",
            "Content-Type": "application/x-www-form-urlencoded",
            "Content-Length": "540",
            "Host": "accounts.nintendo.com",
            "Connection": "Keep-Alive",
            "Accept-Encoding": "gzip",
        }
        params = {
            "client_id": "71b963c1b7b6d119",
            "session_token_code": session_token_code,
            "session_token_code_verifier": self.verifier,
        }
        uri = "https://accounts.nintendo.com/connect/1.0.0/api/session_token"
        response = self.session.post(uri, headers=header, data=params)
        session_token = response.json()["session_token"]
        self._session_token = session_token
        return session_token

    def get_user_access_token(self, session_token: str) -> dict:
        """Obtains the user access token from the session token.

        This method sends a request to Nintendo's servers containing the session
        token to obtain the user access token and the ``id_token``. The user
        access token is then used to obtain the user's data, while the
        ``id_token`` is used to obtain the user's gtoken.

        Args:
            session_token (str): The session token.

        Returns:
            dict: The response from Nintendo's servers. This contains two keys
                of interest: ``access_token`` and ``id_token``. The
                ``access_token`` is used to obtain the user's data, while the
                ``id_token`` is used to obtain the user's gtoken.
        """
        header = {
            "Host": "accounts.nintendo.com",
            "Accept-Encoding": "gzip",
            "Content-Type": "application/json",
            "Content-Length": "436",
            "Accept": "application/json",
            "Connection": "Keep-Alive",
            "User-Agent": "Dalvik/2.1.0 (Linux; U; Android 7.1.2)",
        }
        body = {
            "client_id": "71b963c1b7b6d119",
            "session_token": session_token,
            "grant_type": (
                "urn:ietf:params:oauth:grant-type:jwt-bearer-session-token"
            ),
        }
        uri = "https://accounts.nintendo.com/connect/1.0.0/api/token"
        return self.session.post(uri, headers=header, json=body).json()

    def get_user_info(self, user_access_token: str) -> dict[str, str]:
        """Obtains the user information from the user access token.

        This method will obtain the user information from the user access token.
        This includes the user's set language, country, and birthday, which are
        used to obtain the user's ``gtoken``.

        Args:
            user_access_token (str): The user access token.

        Returns:
            dict[str, str]: The user information. This includes the user's set
                language, country, and birthday.
        """
        # Get user information
        url = "https://api.accounts.nintendo.com/2.0.0/users/me"
        header = {
            "User-Agent": "NASDKAPI; Android",
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"Bearer {user_access_token}",
            "Host": "api.accounts.nintendo.com",
            "Connection": "Keep-Alive",
            "Accept-Encoding": "gzip",
        }
        response = self.session.get(url, headers=header)
        return response.json()

    def get_gtoken(
        self, session_token: str, f_token_url: str | None = None
    ) -> str:
        """Obtains the ``gtoken`` from the session token.

        The GameWebToken, or ``gtoken``, is used to authenticate requests to the
        Nintendo Switch Online API. This method will obtain the ``gtoken`` from
        the session token. The process of obtaining a ``gtoken`` is as follows:

        1. Obtain the user access token and the ``id_token`` from the session
                token.
        2. Use the user access token to obtain the user's information.
        3. Use the ``id_token`` and the user's information to obtain the first
                ``f_token``, the request ID, and the timestamp.
        4. Use the first ``f_token``, the request ID, the timestamp, and the
                ``id_token`` to obtain the ``web_service_access_token``.
        5. Use the ``web_service_access_token`` to obtain the second
                ``f_token``, request ID, and timestamp.
        6. Use the second ``f_token``, request ID, timestamp, and the
                ``web_service_access_token`` to obtain the ``gtoken``.

        By default, this method will use a third party's ``f_token`` generation
        API to obtain the ``f_token``. The API used by default is provided by
        `imink <https://github.com/imink-app>`_. If you do not trust this URL,
        you can provide your own URL through the ``f_token_url`` argument, or
        you can replace the ``f_token`` generation method used with your own
        through the use of the ``set_new_f_token_function`` method. See the
        documentation for that method for more information.

        Args:
            session_token (str): The session token.
            f_token_url (str): The url to get the user access token from. This
                defaults to the ftoken generation url provided by `imink`.

        Raises:
            NintendoException: In the case that the user's access token cannot
                be obtained from the session token, or the user's information
                that is returned is invalid.

        Returns:
            str: The gtoken. This is used to authenticate requests to the
                Nintendo Switch Online API. This token is valid for 2 hours.
        """
        f_token_url = f_token_url if f_token_url is not None else IMINK_URL
        # Get user access token
        logging.info("Getting user access token")
        user_access_response = self.get_user_access_token(session_token)
        try:
            self._user_access_token = cast(
                str, user_access_response["access_token"]
            )
            self._id_token = cast(str, user_access_response["id_token"])
        except (KeyError, TypeError, AttributeError):
            raise NintendoException(
                "Failed to get user access token. "
                + f"Response: {user_access_response}"
            )

        logging.info("Getting user info")
        user_info = self.get_user_info(self._user_access_token)
        self._user_info = user_info
        self._nintendo_account_id = user_info["id"]
        logging.info("Getting Web Service Access Token")
        (
            web_service_access_token,
            coral_user_id,
        ) = self.g_token_generation_phase_1(
            self._id_token,
            user_info,
            self._nintendo_account_id,
            f_token_url=f_token_url,
        )
        logging.info("Getting gtoken")
        self._coral_user_id = coral_user_id
        gtoken = self.g_token_generation_phase_2(
            web_service_access_token,
            self._nintendo_account_id,
            self._coral_user_id,
            f_token_url=f_token_url,
        )
        self._gtoken = gtoken
        return gtoken

    def set_new_f_token_function(
        self, new_function: FToken_Gen | None = None
    ) -> None:
        """Sets the function used to generate the ``f_token``.

        This method will set the function used to generate the ``f_token``. The
        function must take the following arguments:

        1. The ``f_token`` generation url.
        2. The user's token, either the ``id_token`` for step ``1``, or the
            ``web_service_access_token`` for step ``2``.
        3. The step number (either ``1`` or ``2``)

        The function must return a tuple containing the following:

        1. The ``f_token``.
        2. The request id.
        3. The timestamp.

        Args:
            new_function (FToken_Gen): The new function to use to generate the
                ftoken. This function must take the following arguments:

                1. The ``f_token`` generation url.
                2. The user's token, either the ``id_token`` for step ``1``, or
                    the ``web_service_access_token`` for step ``2``.
                3. The step number (either ``1`` or ``2``)

                The function must return a tuple containing the following:

                1. The ``f_token``.
                2. The request id.
                3. The timestamp.

            If this argument is not provided or is ``None``, the default
            ``f_token`` generation method will be restored.
        """
        if new_function is None:
            logging.info("Restoring default ftoken generation method")
            self._f_token_function = self.get_ftoken
        else:
            logging.info("Setting new ftoken generation method")
            self._f_token_function = new_function

    def get_ftoken(
        self,
        f_token_url: str,
        id_token: str,
        step: Literal[1] | Literal[2],
        na_id: str,
        coral_user_id: str | None = None,
    ) -> tuple[str, str, str]:
        """Given the ``f_token_url``, ``id_token``, and ``step``, returns the
        ``f_token``, ``request_id``, and ``timestamp`` from the response.

        Note that this is a third party method, and is not officially
        sanctioned by Nintendo. The default ftoken generation URL used by this
        library is provided by `imink <https://github.com/imink-app>`_. In
        the interest of transparency, the following is the entirety of the
        request header sent to the ftoken generation URL:

        >>> {
        ...     "User-Agent": f"splatnet3_scraper/{__version__}",
        ...     "Content-Type": "application/json; charset=utf-8",
        ... }

        The following is the entirety of the request body sent to the ftoken
        generation URL:

        >>> {
        ...     "token": id_token,
        ...     "hash_method": step,
        ... }

        As you can see, the only identifying information sent to the ftoken
        generation URL is the user's ``id_token``, which cannot be used to
        identify the user without the user's access token, which is not
        provided to the ftoken generation URL.

        Args:
            f_token_url (str): URL to use for ``f_token`` generation. This
                package provides a default URL, but you can provide your own.
                The default URL is provided by `imink`.
            id_token (str): ID token from user access token response. This is
                obtained from the user access token response, and is used to
                identify the user. This cannot be used to identify the user
                without the user's access token, which is not provided to the
                ftoken generation URL.
            step (Literal[1] | Literal[2]): The step number. This is either
                ``1`` or ``2``. This is used to identify the step in the
                ``f_token`` generation process.
            na_id (str): The Nintendo Account ID of the user. As of version
                2.5.1, this is not used for anything. However, it is still
                required to futureproof in case Nintendo decides to enforce
                verification of this value.
            coral_user_id (str | None): The Coral user ID of the user. This is
                used to verify the ftoken generation process. This is only
                required for step ``2``.

        Raises:
            ValueError: In the case that the ``coral_user_id`` is not provided
                for step ``2``.
            FTokenException: In the case that the ftoken cannot be obtained
                from the ftoken generation URL.

        Returns:
            str: The f token.
            str: The request ID.
            str: The timestamp.
        """
        header = {
            "User-Agent": f"splatnet3_scraper/{__version__}",
            "Content-Type": "application/json; charset=utf-8",
            "X-znca-Platform": "Android",
            "X-znca-Version": self.version,
        }
        body = {
            "token": id_token,
            "hash_method": step,
            "na_id": na_id,
        }

        if coral_user_id is not None:
            body["coral_user_id"] = coral_user_id
        elif step == 2:
            raise ValueError(
                "Coral user ID is required for step 2 of ftoken generation"
            )

        response = self.session.post(f_token_url, headers=header, json=body)
        response_json = response.json()
        try:
            f_token = response_json["f"]
            request_id = response_json["request_id"]
            timestamp = response_json["timestamp"]
        except (KeyError, TypeError, AttributeError):
            raise FTokenException(
                "Failed to get f token. " + f"Response: {response_json}"
            )
        return (f_token, request_id, timestamp)

    @retry(times=1, exceptions=(FTokenException, NintendoException, KeyError))
    def g_token_generation_phase_1(
        self,
        id_token: str,
        user_info: dict[str, str],
        na_id: str,
        f_token_url: str,
    ) -> tuple[str, str]:
        """First phase of the ``gtoken`` generation process.

        This is the first phase of the ``gtoken`` generation process. This is
        abstracted away into a separate method to allow the request to be
        retried once in the event of a failure. This phase involves two steps:
        first, the ``id_token`` is sent to the ``f_token_url`` in a request to
        obtain the ``f_token``, which is a ``HMAC`` necessary to obtain the
        ``gtoken``. The response contains three values: the ``f_token``, the
        ``request_id``, and the ``timestamp``. These values are then sent to
        Nintendo's servers along with the  ``id_token`` to obtain the
        ``gtoken``.

        Args:
            id_token (str): ID token from user access token response. This is
                obtained from the user access token response, and is used to
                identify the user to Nintendo's servers. However, this cannot
                be used to identify the user by a third party by itself.
            user_info (dict[str, str]): The dictionary of user info returned by
                ``get_user_info``. This must contain the keys ``language``,
                ``birthday``, and ``country``.
            na_id (str): The Nintendo Account ID of the user. As of version
                2.5.1, this is not used for anything. However, it is still
                required to futureproof in case Nintendo decides to enforce
                verification of this value.
            f_token_url (str): URL to use for f token generation. This package
                provides a default URL, but you can provide your own. The
                default URL is provided by `imink`.

        Returns:
            str: The Web Service Credential Access Token.
            str: The Coral user ID.
        """
        f_token, request_id, timestamp = self._f_token_function(
            f_token_url, id_token, 1, na_id, None
        )
        return self.get_web_service_access_token(
            id_token, user_info, f_token, request_id, timestamp
        )

    @retry(times=1, exceptions=(FTokenException, NintendoException, KeyError))
    def g_token_generation_phase_2(
        self,
        web_service_access_token: str,
        na_id: str,
        coral_user_id: str,
        f_token_url: str,
    ) -> str:
        """Final phase of the ``gtoken`` generation process.

        This is the second phase of the ``gtoken`` generation process. This is
        abstracted away into a separate method to allow the request to be
        retried once in the event of a failure. This phase involves two steps:
        first, the ``web_service_access_token`` is sent to the ``f_token_url``
        in a request to obtain the ``f_token``, which is a ``HMAC`` necessary to
        obtain the ``gtoken``. The response contains three values: the
        ``f_token``, the ``request_id``, and the ``timestamp``. These values are
        then sent to Nintendo's servers along with the
        ``web_service_access_token`` to obtain the ``gtoken``.

        Args:
            web_service_access_token (str): The Web Service Credential Access
                Token obtained from the first half of the gtoken generation
                process.
            na_id (str): The Nintendo Account ID of the user. As of version
                2.5.1, this is not used for anything. However, it is still
                required to futureproof in case Nintendo decides to enforce
                verification of this value.
            coral_user_id (str): The Coral user ID of the user. This is
                used to verify the ftoken generation process. This is only
                required for step ``2``.
            f_token_url (str): URL to use for f token generation. This package
                provides a default URL, but you can provide your own. The
                default URL is provided by `imink`.

        Returns:
            str: The gtoken from the response.
        """
        f_token, request_id, timestamp = self._f_token_function(
            f_token_url, web_service_access_token, 2, na_id, coral_user_id
        )

        return self.get_gtoken_request(
            web_service_access_token, f_token, request_id, timestamp
        )

    def get_web_service_access_token(
        self,
        id_token: str,
        user_info: dict[str, str],
        f_token: str,
        request_id: str,
        timestamp: str,
    ) -> tuple[str, str]:
        """Given the ``id_token``, user data, ``f_token``, ``request_id``, and
        ``timestamp``, returns the Web Service Credential Access Token.

        This is the first step of the ``gtoken`` generation process. This will
        return the Web Service Credential Access Token from Nintendo's servers,
        which is then used to obtain the ``gtoken``.

        Args:
            id_token (str): The ``id_token`` from the user access response.
            user_info (dict[str, str]): The dictionary of user info returned by
                ``get_user_info``. This must contain the keys ``language``,
                ``birthday``, and ``country``.
            f_token (str): The ``f_token`` generated from the ``id_token``. It
                is important that this is the ``f_token`` generated from the
                ``id_token`` using the first hashing method and not the second.
            request_id (str): The ``request_id`` returned alongside the
                ``f_token`` and ``timestamp``.
            timestamp (str): The ``timestamp`` returned alongside the
                ``f_token`` and ``request_id``.

        Returns:
            str: The Web Service Credential Access Token.
            str: The coral user ID.
        """
        header = {
            "X-Platform": "Android",
            "X-ProductVersion": self.version,
            "Content-Type": "application/json; charset=utf-8",
            "Content-Length": str(990 + len(f_token)),
            "Connection": "Keep-Alive",
            "Accept-Encoding": "gzip",
            "User-Agent": "com.nintendo.znca/"
            + self.version
            + "(Android/7.1.2)",
        }
        body = {
            "parameter": {
                "f": f_token,
                "language": user_info["language"],
                "naBirthday": user_info["birthday"],
                "naCountry": user_info["country"],
                "naIdToken": id_token,
                "requestId": request_id,
                "timestamp": timestamp,
            }
        }
        url = "https://api-lp1.znc.srv.nintendo.net/v3/Account/Login"
        response = self.session.post(url, headers=header, json=body).json()
        response_result = response["result"]
        return (
            response_result["webApiServerCredential"]["accessToken"],
            response_result["user"]["id"],
        )

    def get_gtoken_request(
        self,
        web_service_access_token: str,
        f_token: str,
        request_id: str,
        timestamp: str,
    ) -> str:
        """Given the ``web_service_access_token``, ``f_token``, ``request_id``,
        and ``timestamp``, returns the ``gtoken``.

        This is named differently from the other ``get_gtoken`` function to
        avoid confusion. This function specifically makes the request to
        Nintendo's servers to obtain the ``gtoken`` given the web service access
        token, ``f_token``, ``request_id``, and ``timestamp``. The other
        function only takes the ``session_token`` and generates all of the other
        intermediate tokens to obtain the ``gtoken`` by finally calling this
        function.

        Args:
            web_service_access_token (str): The ``web_service_access_token``
                obtained from the first half of the ``gtoken`` generation
                process.
            f_token (str): The ``f_token`` generated from the
                ``web_service_access_token``. It is important that this is the
                ``f_token`` generated from the ``web_service_access_token``
                using the second hashing method and not the first.
            request_id (str): The ``request_id`` returned alongside the
                ``f_token`` and ``timestamp``.
            timestamp (str): The ``timestamp`` returned alongside the
                ``f_token`` and ``request_id``.

        Returns:
            str: The ``gtoken``.
        """
        header = {
            "X-Platform": "Android",
            "X-ProductVersion": self.version,
            "Authorization": f"Bearer {web_service_access_token}",
            "Content-Type": "application/json; charset=utf-8",
            "Content-Length": "391",
            "Accept-Encoding": "gzip",
            "User-Agent": f"com.nintendo.znca/{self.version}(Android/7.1.2)",
        }
        body = {
            "parameter": {
                "f": f_token,
                "id": 4834290508791808,
                "registrationToken": web_service_access_token,
                "requestId": request_id,
                "timestamp": timestamp,
            }
        }
        url = "https://api-lp1.znc.srv.nintendo.net/v2/Game/GetWebServiceToken"
        response = self.session.post(url, headers=header, json=body).json()
        return response["result"]["accessToken"]

    def get_bullet_token(
        self,
        gtoken: str,
        user_info: dict,
        user_agent: str | None = None,
    ) -> str:
        """Given the ``gtoken`` and user information, send a request to SplatNet
        3 to obtain the bullet token. This token is required to make any
        requests to SplatNet 3, and is valid for 6 hours and 30 minutes after it
        is obtained.

        The gtoken is obtained by calling ``get_gtoken`` and the user
        information is obtained by calling ``get_user_info``. The user
        information must contain the keys ``language``, ``birthday``, and
        ``country`` and must align with the values set in the user's Nintendo
        account. The user agent is optional, and if not provided, the default
        user agent will be used. It is not recommended to change the user agent
        from the default unless you know what you are doing.

        Args:
            gtoken (str): GameWebToken, also known as gtoken. Given by Nintendo.
                This token is required to make any requests to Nintendo Switch
                Online services and is valid for 2 hours after it is obtained.
            user_info (dict): Dictionary containing the user's information. This
                must contain the keys ``language``, ``birthday``, and
                ``country`` and must align with the values set in the user's
                Nintendo account. These values are verified by Nintendo, so if
                they do not align, no bullet token will be generated.
            user_agent (str | None): User agent to use for the request. This is
                optional, and if not provided, the default user agent will be
                used. It is not recommended to change the user agent from the
                default unless you know what you are doing. The default user
                agent can be found in ``constants.py`` as
                ``DEFAULT_USER_AGENT``. Defaults to None.

        Raises:
            SplatNetException: Error 401, ``ERROR_INVALID_GAME_WEB_TOKEN``. This
                indicates that the gtoken is invalid. This can happen if the
                gtoken is expired or if the provided gtoken was never valid.
            SplatNetException: Error 403 ``ERROR_OBSOLETE_VERSION``. This
                indicates that the version provided in request header key
                ``X-Web-View-Ver`` is outdated. This can happen if the version
                provided is too old.
            SplatNetException: Error 204, ``USER_NOT_REGISTERED``. This
                indicates that there is no game user associated with Splatoon 3,
                and that the user must play at least one match of Splatoon 3
                before using this library.
            NintendoException: If the request does not fail with one of the
                above errors, this indicates that the request failed for some
                other reason and this exception will be raised to indicate that
                the request failed silently.

        Returns:
            str: The bullet token. This token is required to make any requests
                to SplatNet 3, and is valid for 6 hours and 30 minutes after it
                is obtained.
        """
        logging.info("Getting bullet token")
        user_agent = (
            user_agent if user_agent is not None else DEFAULT_USER_AGENT
        )
        header = {
            "Content-Length": "0",
            "Content-Type": "application/json",
            "Accept-Language": user_info["language"],
            "User-Agent": user_agent,
            "X-Web-View-Ver": self.splatnet_web_version,
            "X-NACOUNTRY": user_info["country"],
            "Accept": "*/*",
            "Origin": SPLATNET_URL,
            "X-Requested-With": "com.nintendo.znca",
        }
        cookies = {
            "_gtoken": gtoken,
            "_dnt": "1",
        }
        url = SPLATNET_URL + "/api/bullet_tokens"
        response = self.session.post(url, headers=header, cookies=cookies)

        if response.status_code == 401:
            raise SplatNetException(
                "Error 401: Invalid Game Web Token (gtoken)"
            )
        elif response.status_code == 403:
            raise SplatNetException("Error 403: Outdated Version")
        elif response.status_code == 204:
            raise SplatNetException(
                "Error 204: User Not Registered. Please play at least one match"
                + " of Splatoon 3 before using this library."
            )

        try:
            return response.json()["bulletToken"]
        except KeyError:
            raise NintendoException("Invalid response from Nintendo")

    @property
    def splatnet_web_version(self) -> str:
        """Get the web view version for SplatNet 3. This is used in the request
        header key ``X-Web-View-Ver`` to indicate the version of the web view
        that is being used. This is required to make any requests to SplatNet
        3. If the version cannot be obtained, a fallback version will be used.

        Returns:
            str: The web view version for SplatNet 3.
        """
        if self._web_view_version is not None:
            return self._web_view_version
        try:
            web_version = get_splatnet_version()
            self._web_view_version = web_version
            return web_version
        except SplatNetException as e:
            logging.warning(str(e))
            logging.warning("Using fallback web view version")
            return WEB_VIEW_VERSION_FALLBACK



from splatnet3_scraper.auth.environment_manager import (
    EnvironmentVariablesManager,
)
from splatnet3_scraper.auth.graph_ql_queries import GraphQLQueries
from splatnet3_scraper.auth.nso import NSO
from splatnet3_scraper.auth.token_manager import Token, TokenManager
